@@ -22,30 +22,17 @@ import socket
 import sys
 import warnings
 
-###############################################################################
-# constants
-###############################################################################
-
-VERSION = (2, 0, 1)
+VERSION = (3, 0, 1)
 HELLO_PREFIX = "OK MPD "
 ERROR_PREFIX = "ACK "
 SUCCESS = "OK"
 NEXT = "list_OK"
 
 
-###############################################################################
-# utils
-###############################################################################
-
-
 def escape(text):
     return text.replace("\\", "\\\\").replace('"', '\\"')
 
 
-###############################################################################
-# logging
-###############################################################################
-
 try:
     from logging import NullHandler
 except ImportError:  # NullHandler was introduced in python2.7
@@ -59,11 +46,6 @@ logger = logging.getLogger(__name__)
 logger.addHandler(NullHandler())
 
 
-###############################################################################
-# exceptions
-###############################################################################
-
-
 class MPDError(Exception):
     pass
 
@@ -92,11 +74,6 @@ class IteratingError(MPDError):
     pass
 
 
-###############################################################################
-# command registration
-###############################################################################
-
-
 class mpd_commands(object):
     """Decorator for registering MPD commands with it's corresponding result
     callback.
@@ -105,6 +82,7 @@ class mpd_commands(object):
     def __init__(self, *commands, **kwargs):
         self.commands = commands
         self.is_direct = kwargs.pop("is_direct", False)
+        self.is_binary = kwargs.pop("is_binary", False)
         if kwargs:
             raise AttributeError(
                 "mpd_commands() got unexpected keyword"
@@ -114,6 +92,7 @@ class mpd_commands(object):
     def __call__(self, ob):
         ob.mpd_commands = self.commands
         ob.mpd_commands_direct = self.is_direct
+        ob.mpd_commands_binary = self.is_binary
         return ob
 
 
@@ -155,11 +134,6 @@ class Noop(object):
     mpd_commands = None
 
 
-###############################################################################
-# abstract base client
-###############################################################################
-
-
 class MPDClientBase(object):
     """Abstract MPD client.
 
@@ -307,6 +281,10 @@ class MPDClientBase(object):
     def _parse_neighbors(self, lines):
         return self._parse_objects_direct(lines, ["neighbor"])
 
+    @mpd_commands("listpartitions", is_direct=True)
+    def _parse_partitions(self, lines):
+        return self._parse_objects_direct(lines, ["partition"])
+
     @mpd_commands(
         "add",
         "addtagid",
@@ -317,6 +295,7 @@ class MPDClientBase(object):
         "crossfade",
         "delete",
         "deleteid",
+        "delpartition",
         "disableoutput",
         "enableoutput",
         "findadd",
@@ -326,8 +305,11 @@ class MPDClientBase(object):
         "mount",
         "move",
         "moveid",
+        "moveoutput",
+        "newpartition",
         "next",
         "outputvolume",
+        "partition",
         "password",
         "pause",
         "ping",
@@ -421,13 +403,9 @@ class MPDClientBase(object):
     def _parse_stickers(self, lines):
         return dict(self._parse_raw_stickers(lines))
 
-    def albumart(self, uri):
-        return self._execute_binary("albumart", [uri])
-
-
-###############################################################################
-# sync client
-###############################################################################
+    @mpd_commands("albumart", "readpicture", is_binary=True)
+    def _parse_plain_binary(self, structure):
+        return structure
 
 
 def _create_callback(self, function, wrap_result):
@@ -479,6 +457,7 @@ class MPDClient(MPDClientBase):
         MPDClientBase._parse_songs,
         MPDClientBase._parse_mounts,
         MPDClientBase._parse_neighbors,
+        MPDClientBase._parse_partitions,
         MPDClientBase._parse_playlists,
         MPDClientBase._parse_database,
         MPDClientBase._parse_messages,
@@ -498,55 +477,14 @@ class MPDClient(MPDClientBase):
 
     def _reset(self):
         super(MPDClient, self)._reset()
-        self._pending = []
         self._iterating = False
         self._sock = None
         self._rbfile = _NotConnected()
         self._wfile = _NotConnected()
 
-    def _send(self, command, args, retval):
-        warnings.warn(
-            "``send_{}`` is deprecated in favor of " "asynchronous API".format(command),
-            DeprecationWarning,
-        )
-        if self._command_list is not None:
-            raise CommandListError(
-                "Cannot use send_{} in a command list".format(command)
-            )
-        self._write_command(command, args)
-        if retval is not None:
-            self._pending.append(command)
-
-    def _fetch(self, command, args, retval):
-        warnings.warn(
-            "``fetch_{}`` is deprecated in favor of "
-            "asynchronous API".format(command),
-            DeprecationWarning,
-        )
-        if self._command_list is not None:
-            raise CommandListError(
-                "Cannot use fetch_{} in a command list".format(command)
-            )
-        if self._iterating:
-            raise IteratingError("Cannot use fetch_{} while iterating".format(command))
-        if not self._pending:
-            raise PendingCommandError("No pending commands to fetch")
-        if self._pending[0] != command:
-            raise PendingCommandError(
-                "'{}' is not the currently pending command".format(command)
-            )
-        del self._pending[0]
-        if callable(retval):
-            return retval()
-        return retval
-
     def _execute(self, command, args, retval):
         if self._iterating:
             raise IteratingError("Cannot execute '{}' while iterating".format(command))
-        if self._pending:
-            raise PendingCommandError(
-                "Cannot execute '{}' with pending commands".format(command)
-            )
         if self._command_list is not None:
             if not callable(retval):
                 raise CommandListError(
@@ -627,73 +565,92 @@ class MPDClient(MPDClientBase):
         return bytes(chunk)
 
     def _read_binary(self):
-        size = None
-        chunk_size = None
-        try:
-            while chunk_size is None:
-                line = self._rbfile.readline().decode("utf-8")
-                if not line.endswith("\n"):
+        """From the data stream, read Unicode lines until one says "binary:
+        <number>\\n"; at that point, read binary data of the given length.
+
+        This behaves like _parse_objects (with empty set of delimiters; even
+        returning only a single result), but rather than feeding from a lines
+        iterable (which would be preprocessed too far), it reads directly off
+        the stream."""
+
+        obj = {}
+
+        while True:
+            line = self._read_line()
+            if line is None:
+                break
+
+            key, value = self._parse_pair(line, ": ")
+
+            if key == "binary":
+                chunk_size = int(value)
+                value = self._read_chunk(chunk_size)
+
+                if len(value) != chunk_size:
+                    self.disconnect()
+                    raise ConnectionError(
+                        "Connection lost while reading binary data: "
+                        "expected %d bytes, got %d" % (chunk_size, len(data))
+                    )
+
+                if self._rbfile.read(1) != b"\n":
+                    # newline after binary content
                     self.disconnect()
                     raise ConnectionError("Connection lost while reading line")
-                line = line.rstrip("\n")
-                if line.startswith(ERROR_PREFIX):
-                    error = line[len(ERROR_PREFIX) :].strip()
-                    raise CommandError(error)
-                field, val = line.split(": ")
-                if field == "size":
-                    size = int(val)
-                elif field == "binary":
-                    chunk_size = int(val)
 
-            if size is None:
-                size = chunk_size
-
-            data = self._read_chunk(chunk_size)
-
-            if len(data) != chunk_size:
-                self.disconnect()
-                raise ConnectionError(
-                    "Connection lost while reading binary data: "
-                    "expected %d bytes, got %d" % (chunk_size, len(data))
-                )
-
-            if self._rbfile.read(1) != b"\n":
-                # newline after binary content
-                self.disconnect()
-                raise ConnectionError("Connection lost while reading line")
-
-            # trailing status indicator
-            # typically OK, but protocol documentation indicates that it is completion code
-            line = self._rbfile.readline().decode("utf-8")
-
-            if not line.endswith("\n"):
-                self.disconnect()
-                raise ConnectionError("Connection lost while reading line")
-
-            line = line.rstrip("\n")
-            if line.startswith(ERROR_PREFIX):
-                error = line[len(ERROR_PREFIX) :].strip()
-                raise CommandError(error)
-
-            return size, data
-        except IOError as err:
-            self.disconnect()
-            raise ConnectionError(
-                "Connection IO error while processing binary command: " + str(err)
-            )
+            obj[key] = value
+        return obj
 
     def _execute_binary(self, command, args):
-        data = bytearray()
+        """Execute a command repeatedly with an additional offset argument,
+        keeping all the identical returned dictionary items and concatenating
+        the binary chunks following the binary item into one of exactly size.
+
+        This differs from _execute in that rather than passing the lines to the
+        callback which'd then call on something like _parse_objects, it builds
+        a parsed object on its own (as a prerequisite to the chunk driving
+        process) and then joins together the chunks into a single big response."""
+        if self._iterating or self._command_list is not None:
+            raise IteratingError("Cannot execute '{}' with command lists".format(command))
+        data = None
+        args = list(args)
         assert len(args) == 1
         args.append(0)
+        final_metadata = None
         while True:
             self._write_command(command, args)
-            size, chunk = self._read_binary()
-            data += chunk
-            args[-1] += len(chunk)
-            if len(data) == size:
+            metadata = self._read_binary()
+            chunk = metadata.pop('binary', None)
+
+            if final_metadata is None:
+                data = chunk
+                final_metadata = metadata
+                if not data:
+                    break
+                try:
+                    size = int(final_metadata['size'])
+                except KeyError:
+                    size = len(chunk)
+                except ValueError:
+                    raise CommandError("Size data unsuitable for binary transfer")
+            else:
+                if metadata != final_metadata:
+                    raise CommandError("Metadata of binary data changed during transfer")
+                if chunk is None:
+                    raise CommandError("Binary field vanished changed during transfer")
+                data += chunk
+            args[-1] = len(data)
+            if len(data) > size:
+                raise CommandListError("Binary data announced size exceeded")
+            elif len(data) == size:
                 break
-        return data
+
+        if data is not None:
+            final_metadata['binary'] = data
+
+        final_metadata.pop('size', None)
+
+        return final_metadata
 
     def _read_command_list(self):
         try:
@@ -826,21 +783,11 @@ class MPDClient(MPDClientBase):
             raise ConnectionError("Not connected")
         return self._sock.fileno()
 
-    def noidle(self):
-        if not self._pending or self._pending[0] != "idle":
-            msg = "cannot send noidle if send_idle was not called"
-            raise CommandError(msg)
-        del self._pending[0]
-        self._write_command("noidle")
-        return self._wrap_iterator(self._parse_list(self._read_lines()))
-
     def command_list_ok_begin(self):
         if self._command_list is not None:
             raise CommandListError("Already in command list")
         if self._iterating:
             raise IteratingError("Cannot begin command list while iterating")
-        if self._pending:
-            raise PendingCommandError("Cannot begin command list with pending commands")
         self._write_command("command_list_ok_begin")
         self._command_list = []
 
@@ -855,15 +802,13 @@ class MPDClient(MPDClientBase):
     @classmethod
     def add_command(cls, name, callback):
         wrap_result = callback in cls._wrap_iterator_parsers
-        method = _create_command(cls._execute, name, callback, wrap_result)
-        send_method = _create_command(cls._send, name, callback, wrap_result)
-        fetch_method = _create_command(cls._fetch, name, callback, wrap_result)
-        # create new mpd commands as function in three flavors:
-        # normal, with "send_"-prefix and with "fetch_"-prefix
+        if callback.mpd_commands_binary:
+            method = lambda self, *args: callback(self, cls._execute_binary(self, name, args))
+        else:
+            method = _create_command(cls._execute, name, callback, wrap_result)
+        # create new mpd commands as function:
         escaped_name = name.replace(" ", "_")
         setattr(cls, escaped_name, method)
-        setattr(cls, "send_" + escaped_name, send_method)
-        setattr(cls, "fetch_" + escaped_name, fetch_method)
 
     @classmethod
     def remove_command(cls, name):
@@ -871,8 +816,6 @@ class MPDClient(MPDClientBase):
             raise ValueError("Can't remove not existent '{}' command".format(name))
         name = name.replace(" ", "_")
         delattr(cls, str(name))
-        delattr(cls, str("send_" + name))
-        delattr(cls, str("fetch_" + name))
 
 
 # vim: set expandtab shiftwidth=4 softtabstop=4 textwidth=79:
